/*-
* Copyright (C) 2005-2006 Nick Withers. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
*
* THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
* OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
* LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
* OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGE.
*/

#include "downtime.h"
#include <err.h>
#include <errno.h>
#include <gtk/gtk.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <sys/param.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>

#ifdef BSD
#include <sysexits.h>
#endif

#include <time.h>

typedef struct
{
	user_specs *specs;
	char *config_path;
} user_specs_and_config_path;

typedef struct
{
	user_specs *specs;
	GtkWidget *gtk_widget;
} user_specs_and_GtkWidget;

static GtkWidget *main_window, *main_time_field, *main_timeout_unit_combo_box;
static GtkWidget *main_shutdown_type_combo_box, *main_toggle_remember, *progress;

static gboolean delete_event_main(void);
static void delete_event_progress(shutdown_t *shutdown);
static void destroy_progress(shutdown_t *shutdown);
static void display_ask_kill_window(shutdown_t *shutdown);
static void display_progress(shutdown_t *shutdown);
static void exit_and_abort(shutdown_t *shutdown);
static void exit_no_abort(void);
static gboolean file_selected_ok(user_specs_and_GtkWidget *params);
static void save_current_user_specs(user_specs_and_config_path *params);
static gboolean pause_shutdown(shutdown_t *shutdown);
static gboolean select_and_open_config(user_specs *specs);
static void show_about_dialog(void);
static gboolean start_shutdown(user_specs *specs);
static gboolean start_shutdown_with_current_specs(user_specs *specs);
static void update_main(user_specs *specs);
static gboolean update_progress(shutdown_t *shutdown);

static void
display_ask_kill_window(shutdown_t *shutdown)
{
	GtkWidget *window, *vbox, *label, *hbox, *button_yes, *button_no;
	char *shutdown_pending_message;

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW (window), "Abort active shutdown?");
	gtk_window_set_resizable(GTK_WINDOW (window), FALSE);
	gtk_window_set_position(GTK_WINDOW (window), GTK_WIN_POS_CENTER);

	gtk_container_set_border_width (GTK_CONTAINER (window), 10);

	vbox = gtk_vbox_new(FALSE, 0);

	if ((pasprintf(&shutdown_pending_message, "There is a shutdown pending (PID %d), would you like to terminate it?", shutdown->pid)) == -1)
		gerrx(EX_OSERR, "pasprintf() returned -1 in create_ask_kill_window(). Error: \"%s\" (%d)", strerror(errno), errno);

	label = gtk_label_new(shutdown_pending_message);

	free(shutdown_pending_message); /* No longer required by the GTK, it stores the string itself */

	hbox = gtk_hbox_new(FALSE, 0);

	button_yes = gtk_button_new_from_stock(GTK_STOCK_YES);
	button_no = gtk_button_new_from_stock(GTK_STOCK_NO);

	gtk_box_pack_start(GTK_BOX (vbox), label, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX (vbox), hbox, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX (hbox), button_yes, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX (hbox), button_no, FALSE, FALSE, 5);
	gtk_container_add(GTK_CONTAINER (window), vbox);

	g_signal_connect_swapped(G_OBJECT (window), "delete_event", G_CALLBACK (exit_and_abort), shutdown);
	g_signal_connect_swapped(G_OBJECT (window), "destroy", G_CALLBACK (exit_and_abort), shutdown);
	g_signal_connect_swapped(G_OBJECT (button_yes), "clicked", G_CALLBACK (exit_and_abort), shutdown);
	g_signal_connect_swapped(G_OBJECT (button_no), "clicked", G_CALLBACK (exit_no_abort), NULL);

	gtk_widget_show(label);
	gtk_widget_show(button_yes);
	gtk_widget_show(button_no);
	gtk_widget_show(hbox);
	gtk_widget_show(vbox);
	gtk_widget_show(window);

	gtk_main();
}

void
display_main_window(user_specs *specs)
{
	extern char *config_path;

	GtkWidget *vbox, *hbox1, *hbox2, *hbox3, *label1, *label2;
	GtkWidget *separator1, *separator2, *button_cancel, *button_ok;
	user_specs_and_config_path specs_and_config_path;

	/*
	* Window proper
	*/

	main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(main_window), PRGNAME);
	gtk_window_set_resizable(GTK_WINDOW(main_window), FALSE);
	gtk_window_set_position(GTK_WINDOW(main_window), GTK_WIN_POS_CENTER);

	vbox = gtk_vbox_new(FALSE, 0);

	/*
	* Menus
	*/

	{
		GtkAccelGroup *accel_group;
		GtkWidget *menu_bar, *file_menu, *help_menu, *file_item, *help_item, *open_menu_item;
		GtkWidget *save_menu_item, *separator_menu_item, *quit_menu_item, *about_menu_item;

		accel_group = gtk_accel_group_new();

		menu_bar = gtk_menu_bar_new();
		gtk_window_add_accel_group(GTK_WINDOW(main_window), accel_group);
		gtk_box_pack_start(GTK_BOX(vbox), menu_bar, FALSE, FALSE, 0);

		file_menu = gtk_menu_new(); /* Menus are not to be gtk_widget_show()n */
		help_menu = gtk_menu_new();

		/* File */
		open_menu_item = gtk_image_menu_item_new_from_stock(GTK_STOCK_OPEN, accel_group);
		save_menu_item = gtk_image_menu_item_new_from_stock(GTK_STOCK_SAVE, accel_group);
		separator_menu_item = gtk_separator_menu_item_new();
		quit_menu_item = gtk_image_menu_item_new_from_stock(GTK_STOCK_QUIT, accel_group);

		specs_and_config_path.specs = specs;
		specs_and_config_path.config_path = config_path;

		g_signal_connect_swapped(G_OBJECT(open_menu_item), "activate", G_CALLBACK(select_and_open_config), specs);
		g_signal_connect_swapped(G_OBJECT(save_menu_item), "activate", G_CALLBACK(save_current_user_specs), &specs_and_config_path);
		g_signal_connect_swapped(G_OBJECT(quit_menu_item), "activate", G_CALLBACK(destroy_main), NULL);

		if (config_path == NULL)
			gtk_widget_set_sensitive(save_menu_item, FALSE);

		gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), open_menu_item);
		gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), save_menu_item);
		gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), separator_menu_item);
		gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), quit_menu_item);

		file_item = gtk_menu_item_new_with_mnemonic("_File");
		gtk_widget_show(file_item);
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(file_item), file_menu);
		gtk_menu_bar_append(GTK_MENU_BAR(menu_bar), file_item);

		/* Help */
		about_menu_item = gtk_image_menu_item_new_from_stock(GTK_STOCK_ABOUT, accel_group);

		g_signal_connect_swapped(G_OBJECT(about_menu_item), "activate", G_CALLBACK(show_about_dialog), NULL);

		gtk_menu_shell_append(GTK_MENU_SHELL(help_menu), about_menu_item);

		help_item = gtk_menu_item_new_with_mnemonic("_Help");
		gtk_widget_show(help_item);
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(help_item), help_menu);
		gtk_menu_bar_append(GTK_MENU_BAR(menu_bar), help_item);

		gtk_widget_show(open_menu_item);
		gtk_widget_show(save_menu_item);
		gtk_widget_show(separator_menu_item);
		gtk_widget_show(quit_menu_item);
		gtk_widget_show(about_menu_item);
		gtk_widget_show(menu_bar);
	}

	/*
	* Timeout selection
	*/

	hbox1 = gtk_hbox_new(FALSE, 0);

	label1 = gtk_label_new("Timeout:");

	main_time_field = gtk_entry_new();

	main_timeout_unit_combo_box = gtk_combo_box_new_text();

	gtk_combo_box_append_text(GTK_COMBO_BOX (main_timeout_unit_combo_box), "minutes"); /* Be careful of order here - replicate in TIMEOUT_TYPES (in downtime.h) */
	gtk_combo_box_append_text(GTK_COMBO_BOX (main_timeout_unit_combo_box), "hours");
	gtk_combo_box_append_text(GTK_COMBO_BOX (main_timeout_unit_combo_box), "days");

	gtk_combo_box_set_active(GTK_COMBO_BOX (main_timeout_unit_combo_box), (gint) specs->timeout_unit);

	/*
	* Shutdown type selection
	*/

	hbox2 = gtk_hbox_new(FALSE, 0);

	label2 = gtk_label_new("Shutdown type:");

	main_shutdown_type_combo_box = gtk_combo_box_new_text();

	gtk_combo_box_append_text(GTK_COMBO_BOX (main_shutdown_type_combo_box), "Power-off"); /* Be careful of order here - replicate in SHUTDOWN_TYPES (in downtime.h) */
	gtk_combo_box_append_text(GTK_COMBO_BOX (main_shutdown_type_combo_box), "Restart");
	gtk_combo_box_append_text(GTK_COMBO_BOX (main_shutdown_type_combo_box), "Halt");
	gtk_combo_box_append_text(GTK_COMBO_BOX (main_shutdown_type_combo_box), "Single user mode");

	gtk_combo_box_set_active(GTK_COMBO_BOX (main_shutdown_type_combo_box), (gint) specs->shutdown_type);

	separator1 = gtk_hseparator_new();

	/*
	* "Remember these settings" and buttons section
	*/

	hbox3 = gtk_hbox_new(FALSE, 0);

	main_toggle_remember = gtk_check_button_new_with_label("Remember these settings");
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (main_toggle_remember), specs->remember);
	separator2 = gtk_hseparator_new();

	if (config_path == NULL) /* If we don't have a configuration file path, we can't save settings */
		gtk_widget_set_sensitive(main_toggle_remember, FALSE);

	button_cancel = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
	button_ok = gtk_button_new_from_stock(GTK_STOCK_OK);

	gtk_box_pack_start(GTK_BOX (hbox1), label1, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX (hbox1), main_time_field, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX (hbox1), main_timeout_unit_combo_box, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX (hbox2), label2, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX (hbox2), main_shutdown_type_combo_box, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX (hbox3), button_ok, FALSE, TRUE, 5);
	gtk_box_pack_start(GTK_BOX (hbox3), button_cancel, FALSE, TRUE, 5);
	gtk_box_pack_start(GTK_BOX (vbox), hbox1, FALSE, TRUE, 5);
	gtk_box_pack_start(GTK_BOX (vbox), hbox2, FALSE, TRUE, 5);
	gtk_box_pack_start(GTK_BOX (vbox), separator1, FALSE, TRUE, 5);
	gtk_box_pack_start(GTK_BOX (vbox), main_toggle_remember, FALSE, TRUE, 5);
	gtk_box_pack_start(GTK_BOX (vbox), separator2, FALSE, TRUE, 5);
	gtk_box_pack_start(GTK_BOX (vbox), hbox3, FALSE, TRUE, 5);

	gtk_container_add(GTK_CONTAINER (main_window), vbox);

	g_signal_connect_swapped(G_OBJECT (main_window), "delete_event", G_CALLBACK (delete_event_main), NULL);
	g_signal_connect_swapped(G_OBJECT (main_window), "destroy", G_CALLBACK (destroy_main), NULL);
	g_signal_connect_swapped(G_OBJECT (main_time_field), "activate", G_CALLBACK (start_shutdown_with_current_specs), specs);
	g_signal_connect_swapped(G_OBJECT (button_cancel), "clicked", G_CALLBACK (destroy_main), NULL);
	g_signal_connect_swapped(G_OBJECT (button_ok), "clicked", G_CALLBACK (start_shutdown_with_current_specs), specs);

	GTK_WIDGET_SET_FLAGS(button_ok, GTK_CAN_DEFAULT);
	gtk_widget_grab_default(button_ok);

	update_main(specs);

	gtk_widget_show(label1);
	gtk_widget_show(main_time_field);
	gtk_widget_show(main_timeout_unit_combo_box);
	gtk_widget_show(label2);
	gtk_widget_show(main_shutdown_type_combo_box);
	gtk_widget_show(separator1);
	gtk_widget_show(main_toggle_remember);
	gtk_widget_show(separator2);
	gtk_widget_show(button_ok);
	gtk_widget_show(button_cancel);
	gtk_widget_show(hbox1);
	gtk_widget_show(hbox2);
	gtk_widget_show(hbox3);
	gtk_widget_show(vbox);
	gtk_widget_show(main_window);

	gtk_main();
}

static gboolean
delete_event_main(void)
{
	gtk_main_quit();

	return (TRUE);
}

static void
delete_event_progress(shutdown_t *shutdown)
{
	if (shutdown->in_progress && shutdown->pid)
		display_ask_kill_window(shutdown);

	exit(EXIT_SUCCESS);
}

gboolean
destroy_main(void)
{
	gtk_main_quit();

	return (FALSE);
}

static void
destroy_progress(shutdown_t *shutdown)
{
	if (shutdown->in_progress && !shutdown->paused && shutdown->pid)
		display_ask_kill_window(shutdown);

	exit(EXIT_SUCCESS);
}

static void
display_progress(shutdown_t *shutdown)
{
	char *str_time_to_shutdown;
	GtkWidget *window, *vbox, *hbox, *button_close, *pause_button;

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW (window), "Shutdown Progress - " PRGNAME);
	gtk_window_set_resizable(GTK_WINDOW (window), FALSE);
	gtk_window_set_position(GTK_WINDOW (window), GTK_WIN_POS_CENTER);
	gtk_container_set_border_width (GTK_CONTAINER (window), 10);

	vbox = gtk_vbox_new(FALSE, 0);

	progress = gtk_progress_bar_new();
	gtk_progress_bar_set_text(GTK_PROGRESS_BAR (progress), (str_time_to_shutdown = get_shutdown_time_left_string(shutdown)));
	g_timeout_add((((shutdown->shutdown_time - shutdown->start_time) > 100) ? 1000 : (shutdown->shutdown_time - shutdown->start_time) * 10), (GSourceFunc) update_progress, shutdown); /* Read "((shutdown->shutdown_time - shutdown->start_time) > 100) ? 1000" as "(shutdown->shutdown_time - shutdown->start_time) * 10 > 1000". * 10 because the first part of the expression should be divided by 100 to get a percentage, then multiplied by 1000 to go from seconds to milliseconds */
	free(str_time_to_shutdown);

	hbox = gtk_hbox_new(FALSE, 0);

	pause_button = gtk_button_new_from_stock(GTK_STOCK_MEDIA_PAUSE);
	button_close = gtk_button_new_from_stock(GTK_STOCK_CLOSE);

	gtk_box_pack_start(GTK_BOX (vbox), progress, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX (hbox), pause_button, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX (hbox), button_close, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX (vbox), hbox, FALSE, FALSE, 5);
	gtk_container_add(GTK_CONTAINER (window), vbox);

	g_signal_connect_swapped(G_OBJECT (window), "delete_event", G_CALLBACK (delete_event_progress), shutdown);
	g_signal_connect_swapped(G_OBJECT (window), "destroy", G_CALLBACK (destroy_progress), shutdown);
	g_signal_connect_swapped(G_OBJECT (pause_button), "clicked", G_CALLBACK (pause_shutdown), shutdown);
	g_signal_connect_swapped(G_OBJECT (button_close), "clicked", G_CALLBACK (destroy_progress), shutdown);

	if (!shutdown->pid)
		gtk_widget_set_sensitive(pause_button, FALSE); /* Disable the pause button if we don't know the shutdown(8) process (if any)' PID */

	gtk_widget_show(pause_button);
	gtk_widget_show(button_close);
	gtk_widget_show(progress);
	gtk_widget_show(hbox);
	gtk_widget_show(vbox);
	gtk_widget_show(window);

	gtk_main();
}

static void
exit_and_abort(shutdown_t *shutdown)
{
	if (kill(shutdown->pid, SIGTERM) != 0) /* shutdown->pid is checked to ensure it's > 0 in shutdown.c */
		gerrx(EX_OSERR, "Couldn't kill PID %d, error: \"%s\" (%d).\n\nPlease kill the shutdown(8) process manually.", shutdown->pid, strerror(errno), errno);

	exit(EXIT_SUCCESS);
}

static void
exit_no_abort(void)
{
	exit(EXIT_SUCCESS);
}

static gboolean
file_selected_ok(user_specs_and_GtkWidget *params)
{
	char *config_path;

	config_path = (char *) gtk_file_selection_get_filename(GTK_FILE_SELECTION (params->gtk_widget));

	if (load_user_specs(params->specs, config_path)) /* Not doing any checking of config_path here. If for any reason it's invalid, load_settings() should spit */
		update_main(params->specs);
	else
		gwarnx("Unable to open configuration file \"%s\". Error: \"%s\" (%d)", config_path, strerror(errno), errno);

	return (FALSE);
}

void
gerrx(int eval, const char *fmt, ...)
{
	va_list ap;
	char *exp_fmt;
	GtkWidget *message_dialog;

	va_start(ap, fmt);

	if ((pvasprintf(&exp_fmt, fmt, ap)) == -1)
		errx(EX_OSERR, "pvasprintf() returned -1 in gerrx(). Error: \"%s\" (%d)", strerror(errno), errno);

	va_end(ap);

	warnx(exp_fmt);

	message_dialog = gtk_message_dialog_new(NULL, (GtkDialogFlags) NULL, GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE, exp_fmt);
	gtk_dialog_run(GTK_DIALOG (message_dialog));
	gtk_widget_destroy(message_dialog);

	exit(eval);
}

/* Updates the structure pointed to by specs with user-supplied specifications from the DownTime main dialogue. DO NOT call after the main window has been destroyed */
char
update_user_specs(user_specs *specs)
{
	char *nptr, *endptr;
	long int timeout;

	nptr = (char *) gtk_entry_get_text(GTK_ENTRY (main_time_field));
	timeout = (unsigned long) strtol(nptr, &endptr, 10);

	if (!(*nptr != (char) NULL && *endptr == (char) NULL))
	{
		gwarnx("An invalid timeout - \"%s\" - was specified", nptr);

		return (FALSE);
	}

	specs->timeout = timeout;
	specs->timeout_unit = gtk_combo_box_get_active(GTK_COMBO_BOX (main_timeout_unit_combo_box));
	specs->shutdown_type = gtk_combo_box_get_active(GTK_COMBO_BOX (main_shutdown_type_combo_box));
	specs->remember = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (main_toggle_remember));

	return (TRUE);
}

void
gwarnx(const char *fmt, ...)
{
	va_list ap;
	char *exp_fmt;
	GtkWidget *message_dialog;

	va_start(ap, fmt);

	if ((pvasprintf(&exp_fmt, fmt, ap)) == -1)
		gerrx(EX_OSERR, "pvasprintf() returned -1 in gerrx(). Error: \"%s\" (%d)", strerror(errno), errno);

	va_end(ap);

	warnx(exp_fmt);

	message_dialog = gtk_message_dialog_new(NULL, (GtkDialogFlags) NULL, GTK_MESSAGE_WARNING, GTK_BUTTONS_OK, exp_fmt);
	gtk_dialog_run(GTK_DIALOG (message_dialog));
	gtk_widget_destroy(message_dialog);

	free(exp_fmt);
}

static gboolean
pause_shutdown(shutdown_t *shutdown)
{
	extern char *shutdown_path;

	time_t current_time;

	current_time = time(NULL);

	if (!shutdown->paused) /* The shutdown isn't currently paused */
	{
		if (kill(shutdown->pid, SIGTERM) != 0) /* We're not checking that the PID is valid here as this has already been done in launch_shutdown(), which must have been successfully called if we're here */
		{
			gwarnx("Couldn't kill PID %d in order to pause the shutdown. Error: \"%s\" (%d)", shutdown->pid, strerror(errno), errno);

			return (FALSE);
		}

		shutdown->timeout = max_minutes_to_shutdown(shutdown);
		shutdown->start_time -= current_time; /* shutdown->start_time now holds the start time relative to the current time (<= 0), rather than relative to the epoch */
		shutdown->paused = TRUE; /* Indicate that a shutdown is paused. This'll stop the progress bar from updating */

		gtk_progress_bar_set_text(GTK_PROGRESS_BAR (progress), "(Paused)");
	}
	else
	{
		user_specs specs;
		char *str_time_to_shutdown;

		shutdown->start_time += current_time;
		shutdown->shutdown_time = current_time + (shutdown->timeout * 60);
		specs.timeout = shutdown->timeout;
		specs.timeout_unit = MINUTES;
		specs.shutdown_type = shutdown->type;
		specs.remember = FALSE;

		if (launch_shutdown(shutdown_path, &specs, shutdown) == NULL)
			gerrx(EX_OSERR, "Couldn't re-launch shutdown(8) following a pause. Error: \"%s\" (%d)", strerror(errno), errno);

		shutdown->paused = FALSE; /* Indicate that no shutdown is paused, allowing the progress bar to start updating again */

		gtk_progress_bar_set_text(GTK_PROGRESS_BAR (progress), str_time_to_shutdown = get_shutdown_time_left_string(shutdown));

		free(str_time_to_shutdown); /* The GTK doesn't need this anymore, as it copies the text to its own memory area */
	}

	return (TRUE);
}

static void
save_current_user_specs(user_specs_and_config_path *params)
{
	if (!update_user_specs(params->specs))
		return;

	save_user_specs(params->specs, params->config_path);
}

static gboolean
select_and_open_config(user_specs *specs)
{
	extern char *config_path;

	GtkWidget *file_selection;
	user_specs_and_GtkWidget specs_and_widget;

	file_selection = gtk_file_selection_new("Select configuration file");
	gtk_file_selection_set_filename(GTK_FILE_SELECTION (file_selection), config_path == NULL ? CONFIG_FILE_NAME : config_path);

	specs_and_widget.specs = specs;
	specs_and_widget.gtk_widget = file_selection;

	g_signal_connect_swapped(G_OBJECT (file_selection), "destroy", G_CALLBACK (delete_event_main), NULL);
	g_signal_connect_swapped(G_OBJECT (GTK_FILE_SELECTION (file_selection)->ok_button), "clicked", G_CALLBACK (file_selected_ok), &specs_and_widget); /* It's OK to pass &specs_and_widget here, since it'll always be valid when file_selected_ok is called (gtk_main() is called in this function) */
	g_signal_connect_swapped(G_OBJECT (GTK_FILE_SELECTION (file_selection)->ok_button), "clicked", G_CALLBACK (gtk_widget_destroy), file_selection);
	g_signal_connect_swapped(G_OBJECT (GTK_FILE_SELECTION (file_selection)->cancel_button), "clicked", G_CALLBACK (gtk_widget_destroy), G_OBJECT (file_selection));

	gtk_widget_show(GTK_WIDGET(file_selection));

	gtk_main();

	return (TRUE);
}

static void
show_about_dialog(void)
{
	GtkWidget *about_dialog;
	char *authors[] = AUTHORS;

	about_dialog = gtk_about_dialog_new();

	gtk_about_dialog_set_name(GTK_ABOUT_DIALOG (about_dialog), PRGNAME);
	gtk_about_dialog_set_version(GTK_ABOUT_DIALOG (about_dialog), VERSION);
	gtk_about_dialog_set_copyright(GTK_ABOUT_DIALOG (about_dialog), COPYRIGHT);
	gtk_about_dialog_set_website(GTK_ABOUT_DIALOG (about_dialog), CONTACT_WWW);
	gtk_about_dialog_set_license(GTK_ABOUT_DIALOG (about_dialog), LICENSE_LONG);
	gtk_about_dialog_set_authors(GTK_ABOUT_DIALOG (about_dialog), (const gchar **) authors);

	gtk_widget_show(about_dialog);
}

static gboolean
start_shutdown(user_specs *specs)
{
	extern char *config_path, *shutdown_path;

	shutdown_t *shutdown;

	if (specs->remember)
		save_user_specs(specs, config_path);

	if ((shutdown = malloc(sizeof (*shutdown))) == NULL)
		gerrx(EX_OSERR, "malloc() returned NULL in start_shutdown(). Error: \"%s\" (%d)", strerror(errno), errno);

	if (launch_shutdown(shutdown_path, specs, shutdown) == NULL)
		return (FALSE);

	gtk_widget_hide_all(main_window); /* Hide the main window. Actually want to kill it, but doing so at this point makes gtk_main() (called in main()) return, hence terminating DownTime */
	gtk_widget_unrealize(main_window); /* This may well be naughty... */

	display_progress(shutdown);

	return (TRUE);
}

static gboolean
start_shutdown_with_current_specs(user_specs *specs)
{
	if (!update_user_specs(specs)) /* If we weren't able to successfully read the user specs... Note also that update_user_specs() deals with alerting the user about input dramas */
		return (FALSE);

	return (start_shutdown(specs));
}

static void
update_main(user_specs *specs)
{
	char *timeout;

	if ((pasprintf(&timeout, "%lu", specs->timeout)) == -1)
		gerrx(EX_OSERR, "pasprintf() returned -1 in update_main(). Error: \"%s\" (%d)", strerror(errno), errno);

	gtk_entry_set_text(GTK_ENTRY(main_time_field), timeout);

	free(timeout);

	gtk_combo_box_set_active(GTK_COMBO_BOX (main_timeout_unit_combo_box), (gint) specs->timeout_unit);
	gtk_combo_box_set_active(GTK_COMBO_BOX (main_shutdown_type_combo_box), (gint) specs->shutdown_type);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (main_toggle_remember), specs->remember);
}

static gboolean
update_progress(shutdown_t *shutdown)
{
	double new_progress_percent;
	char *str_time_to_shutdown;

	if (shutdown->paused) /* If we're paused, we only need to update the progress bar text */
		return (TRUE); /* Return TRUE so that the timer call to this function isn't canned */

	if (!shutdown->in_progress || progress == NULL)
		return (FALSE);

	if (shutdown->pid && kill(shutdown->pid, 0) != 0) /* Ensure that shutdown(8) is still running. shutdown->pid is checked to ensure it's > 0 in shutdown.c */
	{
		int status;

		waitpid(shutdown->pid, &status, WNOHANG);

		if (WIFEXITED (status))
			gerrx(EX_UNAVAILABLE, "shutdown(8) (PID: %ld) exited unexpectedly with exit status %d (\"%s\")", shutdown->pid, WEXITSTATUS (status), strerror(WEXITSTATUS (status)));
		else
			gerrx(EX_UNAVAILABLE, "shutdown(8) (PID: %ld) terminated unexpectedly", shutdown->pid);
	}

	new_progress_percent = difftime(shutdown->shutdown_time, shutdown->start_time) > 0 ? (100 * ((double) difftime(time(NULL), shutdown->start_time)) / (double) (difftime(shutdown->shutdown_time, shutdown->start_time))) : 0;

	if (new_progress_percent > 100)
		new_progress_percent = 100;
	else if (new_progress_percent < 0)
		new_progress_percent = 0;

	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR (progress), new_progress_percent / 100);
	gtk_progress_bar_set_text(GTK_PROGRESS_BAR (progress), (str_time_to_shutdown = get_shutdown_time_left_string(shutdown)));

	free(str_time_to_shutdown);

	return (TRUE);
}
